"""Add a nested Table of Contents to any HTML file with BeautifulSoup and dominate."""

from bs4 import BeautifulSoup
from dominate import tags

TOC_KEYWORD = '{{toc}}'
"""Default string to replace in the specified file with the nested table of contents. Default is `{{toc}}`."""


def add_nested_list_item(l_index, l_string, level=1):
    """Add nested list items recursively.

    Args:
        l_index: numeric index of the list depth (note: 1-indexed)
        l_string: string to show in the list element
        level: current list depth. Optional and default is 1

    """
    with tags.ul():
        if l_index != level:
            add_nested_list_item(l_index, l_string, level + 1)
        else:
            tags.li(f'H{l_index}: {l_string}')


def create_toc(html_text, header_depth=3):
    """Return the HTML for a nested table of contents based on the HTML file path.

    Args:
        html_text: HTML text
        header_depth: depth of headers to show. Default is 3 (H1-H3)

    Returns:
        string: table of contents

    """
    soup = BeautifulSoup(html_text, features='lxml')
    h_lookup = {f'h{idx}': idx for idx in range(1, header_depth + 1)}
    toc = tags.div()
    for header in soup.findAll([*h_lookup.keys()]):
        with toc:
            add_nested_list_item(h_lookup[header.name], header.string)
            # FIXME: Figure out how to make the header links work (i.e. when clicked in TOC go to the respective header)
            # > `tags.a(header.string, f'#{header.string}')`?
    return str(toc)


def write_toc(html_path, header_depth=3, toc_key=TOC_KEYWORD):
    """Write the nested table of contents to the specified file.

    Args:
        html_path: path to the HTML file
        header_depth: depth of headers to show. Default is 3 (H1-H3)
        toc_key: string to replace with the nested table of contents. Default is `TOC_KEYWORD`

    Raises:
        RuntimeError: if the key was not found in the file

    """
    text = html_path.read_text()
    if toc_key not in text:
        raise RuntimeError(f'HTML file does not have the table of contents key ({toc_key}): {html_path}')
    toc = create_toc(text, header_depth=header_depth)
    html_path.write_text(text.replace('{{toc}}', toc))



"""Dash Modules for Data Tables."""

import json

import dash_bootstrap_components as dbc
import pandas as pd
from dash import dcc, html
from dash.exceptions import PreventUpdate

from .components import dropdown_group, opts_dd
from .datatable import BaseDataTable
from .utils_app_modules import ModuleBase
from .utils_callbacks import map_args, map_outputs


class ModuleDataTable(ModuleBase):
    """Modular Dash data table."""

    id_table_parent = 'datatable-module-parent'
    """Table Parent ID."""

    id_table = 'datatable-module'
    """Table ID."""

    table = None
    """Dash Data Table."""

    all_ids = [id_table_parent, id_table]
    """List of ids to register for this module."""

    def create_elements(self, ids):
        """Register the callback for creating the main chart.

        Args:
            ids: `self._il` from base application

        """
        self.table = BaseDataTable()

    def return_layout(self, ids):
        """Return Dash application layout.

        Args:
            ids: `self._il` from base application

        Returns:
            dict: Dash HTML object

        """
        placeholder = pd.DataFrame.from_records([['body']], columns=['header'])
        return html.Div(
            [
                self.table.create_table(placeholder, None, id=ids[self.get(self.id_table)]),
            ], id=ids[self.get(self.id_table_parent)],
        )

    def return_table_map(self, ids, df_table, columns=None):
        """Return list of tuples for `map_outputs` that includes the new datatable.

        Args:
            ids: `self._il` from base application
            df_table: dataframe to show in table
            columns: list of columns to show. Default is None, which will show all columns

        Returns:
            list: list of tuples for `map_outputs`

        """
        datatable = self.table.create_table(df_table, columns, id=ids[self.get(self.id_table)])
        return [(self.get(self.id_table_parent), 'children', datatable)]

    def create_callbacks(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        self.register_highlight_sort_column(parent)

    def register_highlight_sort_column(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        Raises:
            PreventUpdate: if no columns found in the table

        """
        outputs = [(self.get(self.id_table), 'style_data_conditional')]
        inputs = [(self.get(self.id_table), 'sort_by')]
        states = []

        @parent.callback(outputs, inputs, states)
        def highlight_sort_column(*raw_args):
            a_in, a_states = map_args(raw_args, inputs, states)
            sort_by = a_in[self.get(self.id_table)]['sort_by']
            if sort_by is None:
                raise PreventUpdate

            sorted_columns_style = []
            for sort in sort_by:
                sorted_columns_style.extend([
                    {
                        'if': {'column_id': sort['column_id'], 'row_index': 'odd'},
                        'color': self.table.text_color, 'background-color': self.table.zebra_color,
                    },
                    {
                        'if': {'column_id': sort['column_id'], 'row_index': 'even'},
                        'color': self.table.text_color, 'background-color': self.table.selected_cell_color,
                    },
                ])

            style_data_conditional = [*self.table.style_data_conditional, *sorted_columns_style]
            return map_outputs(outputs, [(self.get(self.id_table), 'style_data_conditional', style_data_conditional)])


class ModuleFilteredTable(ModuleDataTable):
    """Modular Dash data table with column selection and filter."""

    id_column_select = 'column-select'
    """Column select ID."""

    id_filter_button = 'filter-button'
    """Apply filter button ID."""

    id_filter_input = 'filter-query-input'
    """Filter query input ID."""

    id_filter_output = 'filter-query-output'
    """Filter query output ID."""

    id_filter_structure = 'filter-query-structure'
    """Filter query structure ID."""

    all_ids = ModuleDataTable.all_ids + [
        id_column_select, id_filter_button, id_filter_input, id_filter_output,
        id_filter_structure,
    ]
    """List of ids to register for this module."""

    show_filter = True
    """If True (default), will show an input for entering a global filter."""

    mod_df = None
    """Data frame shown in table. Passed to `return_layout` and used when creating the table."""

    def return_layout(self, ids, mod_df):
        """Return Dash application layout.

        Args:
            ids: `self._il` from base application
            mod_df: dataframe for

        Returns:
            dict: Dash HTML object

        """
        self.mod_df = mod_df
        options = [opts_dd(column, column) for column in self.mod_df.columns]
        if self.show_filter:
            filter_elements = dbc.Row([
                dbc.Col(
                    [
                        dbc.Form([
                            dcc.Input(
                                id=ids[self.get(self.id_filter_input)], placeholder='Enter filter query',
                                style={'width': '100%'},
                            ),
                            dbc.Button(
                                'Apply', color='secondary', id=ids[self.get(self.id_filter_button)],
                                style={'paddingTop': '5px'},
                            ),
                        ]),
                    ], width=4,
                ),
                dbc.Col(
                    [
                        html.Div([], id=ids[self.get(self.id_filter_output)]),
                    ], width=4,
                ),
                dbc.Col(
                    [
                        html.Div(id=ids[self.get(self.id_filter_structure)], style={'whitespace': 'pre'}),
                    ], width=4,
                ),
            ])
        else:
            filter_elements = html.Div()

        return dbc.Col([
            dropdown_group(
                'Select DataFrame Columns', ids[self.get(self.id_column_select)],
                options, multi=True, persistence=True, value=self.mod_df.columns,
            ),
            filter_elements,
            html.Br(),
            super().return_layout(ids),
        ])

    def create_callbacks(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        super().create_callbacks(parent)
        self.register_create_table(parent)
        if self.show_filter:
            self.register_filter_interface(parent)
            self.register_show_query(parent)

    def register_create_table(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        Raises:
            PreventUpdate: if no columns found in the table

        """
        outputs = [(self.get(self.id_table_parent), 'children')]
        inputs = [(self.get(self.id_column_select), 'value')]
        states = []

        @parent.callback(outputs, inputs, states)
        def create_table(*raw_args):
            a_in, a_states = map_args(raw_args, inputs, states)
            columns = a_in[self.get(self.id_column_select)]['value']
            if not columns:
                raise PreventUpdate
            return map_outputs(outputs, self.return_table_map(parent.ids, self.mod_df, columns))

    def register_filter_interface(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        outputs = [(self.get(self.id_table), 'filter_query')]
        inputs = [(self.get(self.id_filter_button), 'n_clicks')]
        states = [(self.get(self.id_filter_input), 'value')]

        @parent.callback(outputs, inputs, states)
        def show_query(*raw_args):
            a_in, a_states = map_args(raw_args, inputs, states)
            query = a_states[self.get(self.id_filter_input)]['value']
            if query is None:
                query = ''
            return map_outputs(outputs, [(self.get(self.id_table), 'filter_query', query)])

    def register_show_query(self, parent):   # noqa: CCR001
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        outputs = [(self.get(self.id_filter_output), 'children'), (self.get(self.id_filter_structure), 'children')]
        inputs = [
            (self.get(self.id_table), 'filter_query'),
            (self.get(self.id_table), 'derived_filter_query_structure'),
        ]
        states = []

        @parent.callback(outputs, inputs, states)
        def show_query(*raw_args):
            a_in, a_states = map_args(raw_args, inputs, states)
            filter_query = a_in[self.get(self.id_table)]['filter_query']
            derived_query = a_in[self.get(self.id_table)]['derived_filter_query_structure']

            if filter_query is None or len(filter_query.strip()) == 0:
                filter_element = ['No filter query']
                derived_element = ['']
            else:
                filter_element = [dcc.Markdown(f'```\n\nfilter_query:\n\n{filter_query}\n\n```')]
                if derived_query is None:
                    derived_element = ['Error in query. Check formatting']
                else:
                    derived_element = [
                        html.Details([
                            html.Summary('Derived filter query structure'),
                            html.Div(dcc.Markdown(f'```json\n\n{json.dumps(derived_query, indent=4)}\n\n```')),
                        ]),
                    ]
            return map_outputs(
                outputs, [
                    (self.get(self.id_filter_output), 'children', filter_element),
                    (self.get(self.id_filter_structure), 'children', derived_element),
                ],
            )



"""Pareto Chart."""

import pandas as pd
import plotly.graph_objects as go

from .utils_data import append_df, validate
from .utils_fig import CustomChart, check_raw_data


def tidy_pareto_data(df_raw, cap_categories):
    """Return compressed Pareto dataframe of only the unique values.

    Args:
        df_raw: pandas dataframe with at minimum the two columns `category: str` and `value: float`
        cap_categories: Maximum number of categories (bars)

    Returns:
        dataframe: pandas dataframe with columns `(value, label, counts, cum_per)`

    """
    df_p = None
    for cat in df_raw['category'].unique():
        df_row = pd.DataFrame(
            data={
                'label': [cat],
                'value': [df_raw.loc[df_raw['category'] == cat]['value'].sum()],
                'counts': df_raw['category'].value_counts()[cat],
            },
        )
        df_p = append_df(df_p, df_row)
    # Sort and calculate percentage
    df_p = (
        df_p[df_p['value'] != 0]
        .sort_values(by=['value'], ascending=False)
        .head(cap_categories)
    )
    df_p['cum_per'] = df_p['value'].divide(df_p['value'].sum()).cumsum()
    return df_p


class ParetoChart(CustomChart):
    """Pareto Chart: both bar and line graph chart for strategic decision making."""

    cap_categories: int = 20
    """Maximum number of categories (bars). Default is 20."""

    show_count: bool = True
    """If True, will show numeric count on each bar. Default is True."""

    yaxis_2_label: str = 'Cumulative Percentage'
    """Label for yaxis 2 that shows the cumulative percentage."""

    _pareto_colors: dict = {'bar': '#4682b4', 'line': '#b44646'}
    _pareto_colors_schema = {
        'bar': {'required': True, 'type': 'string'},
        'line': {'required': True, 'type': 'string'},
    }

    @property
    def pareto_colors(self):
        """Colors for bar and line traces in Pareto chart.

        Returns:
            dict: dictionary with keys `(bar, line)`

        """
        return self._pareto_colors

    @pareto_colors.setter
    def pareto_colors(self, pareto_colors):
        errors = validate(pareto_colors, self._pareto_colors_schema)
        if errors:
            raise RuntimeError(f'Validation of self.pareto_colors failed: {errors}')
        # Assign new pareto_colors
        self._pareto_colors = pareto_colors

    def create_traces(self, df_raw):
        """Return traces for plotly chart.

        Args:
            df_raw: pandas dataframe with at minimum the two columns `category: str` and `value: float`

        Returns:
            list: Dash chart traces

        Raises:
            RuntimeError: if the `df_raw` is missing any necessary columns

        """
        # Check that the raw data frame is properly formatted
        check_raw_data(df_raw, min_keys=['category', 'value'])
        if not pd.api.types.is_string_dtype(df_raw['category']):  # pragma: no cover
            raise RuntimeError(f"category column must be string, but found {df_raw['category'].dtype}")

        # Create and return the traces and optionally add the count to the bar chart
        df_p = tidy_pareto_data(df_raw, self.cap_categories)
        count_kwargs = {'text': df_p['counts'], 'textposition': 'auto'} if self.show_count else {}
        return [
            go.Bar(
                hoverinfo='y', yaxis='y1', name='raw_value',
                marker={'color': self.pareto_colors['bar']},
                x=df_p['label'], y=df_p['value'], **count_kwargs,
            ),
        ] + [
            go.Scatter(
                hoverinfo='y', yaxis='y2', name='cumulative_percentage',
                marker={'color': self.pareto_colors['line']}, mode='lines',
                x=df_p['label'], y=df_p['cum_per'],
            ),
        ]

    def create_layout(self):
        """Extend the standard layout.

        Returns:
            dict: layout for Dash figure

        """
        layout = super().create_layout()
        layout['legend'] = {}
        layout['showlegend'] = False

        layout['margin'] = {'l': 75, 'b': 100, 't': 50, 'r': 125}

        # Update YAxis configuration
        layout['yaxis']['mirror'] = 'ticks'
        layout['yaxis']['showline'] = True
        layout['yaxis']['tickformat'] = '.0f'

        # See multiple axis: https://plot.ly/python/multiple-axes/
        layout['yaxis2'] = {
            'dtick': 0.1,
            'overlaying': 'y',
            'range': [0, 1.01],
            'showgrid': False,
            'side': 'right',
            'tickformat': '.0%',
            'tickmode': 'linear',
            'title': self.yaxis_2_label,
        }

        return layout



"""Custom Plotly Colorscales."""

DEFAULT_PLOTLY_COLORS = [
    '#1f77b4',  # muted blue
    '#ff7f0e',  # safety orange
    '#2ca02c',  # cooked asparagus green
    '#d62728',  # brick red
    '#9467bd',  # muted purple
    '#8c564b',  # chestnut brown
    '#e377c2',  # raspberry yogurt pink
    '#7f7f7f',  # middle gray
    '#bcbd22',  # curry yellow-green
    '#17becf',  # blue-teal
]
"""List of default Plotly colors in Hex strings."""

DEFAULT_PLOTLY_COLORS_RGB = [
    'rgb(31,119,180)',   # 0
    'rgb(255,127,14)',   # 1
    'rgb(44,160,44)',    # 2
    'rgb(214,39,40)',    # 3
    'rgb(148,103,189)',  # 4
    'rgb(140,86,75)',    # 5
    'rgb(227,119,194)',  # 6
    'rgb(127,127,127)',  # 7
    'rgb(188,189,34)',   # 8
    'rgb(23,190,207)',   # 9
]
"""List of default Plotly colors in RGB strings."""

# From SF Example: (#fdae61, #abd9e9, #2c7bb6)

# Plotly Colors:
# ['Blackbody', 'Blackbody_r', 'Bluered', 'Bluered_r', 'Blues', 'Blues_r', 'Cividis', 'Cividis_r', 'Earth', 'Earth_r',
#  'Electric', 'Electric_r', 'Greens', 'Greens_r', 'Greys', 'Greys_r', 'Hot', 'Hot_r', 'Jet', 'Jet_r', 'Picnic',
#  'Picnic_r', 'Portland', 'Portland_r', 'Rainbow', 'Rainbow_r', 'RdBu', 'RdBu_r', 'Reds', 'Reds_r', 'Viridis',
#  'Viridis_r', 'YlGnBu', 'YlGnBu_r', 'YlOrRd', 'YlOrRd_r', 'scale_name', 'scale_name_r', 'scale_pairs',
#  'scale_pairs_r', 'scale_sequence', 'scale_sequence_r']
# plotly.colors.plotlyjs.Hot / `[rgb(0,0,0), rgb(230,0,0), rgb(255,210,0), rgb(255,255,255)]`



"""Utilities for custom Dash figures."""

import plotly.graph_objects as go
from dash import dcc
from plotly.subplots import make_subplots

from .utils_data import validate

FIGURE_PLACEHOLDER = {'data': [], 'layout': {}, 'frames': []}
"""Figure placeholder."""


def min_graph(config=None, figure=FIGURE_PLACEHOLDER, **kwargs):
    """Return dcc.Graph element with Plotly overlay removed.

    See: https://community.plot.ly/t/is-it-possible-to-hide-the-floating-toolbar/4911/7

    Args:
        config: dictionary passed to `dcc.Graph`. Default is to disable the `displayModeBar`
        figure: figure argument. Default is an empty placeholder (`FIGURE_PLACEHOLDER`)
        kwargs: any kwargs to pass to the dash initializer other than `assets_folder`

    Returns:
        dict: Dash `dcc.Graph` object

    """
    if config is None:
        config = {'displayModeBar': False}
    return dcc.Graph(config=config, figure=figure, **kwargs)


def check_raw_data(df_raw, min_keys):
    """Verify that dataframe contains the minimum columns.

    Args:
        df_raw: data to pass to formatter method
        min_keys: list of string column names expected in dataframe

    Raises:
        RuntimeError: if any columns are missing from dataframe

    """
    all_keys = df_raw.keys()
    if len([_k for _k in min_keys if _k in all_keys]) != len(min_keys):
        raise RuntimeError(f'`df_raw` must have keys {min_keys}. Found: {all_keys}')


def make_dict_an(coord, text, label=None, color=None, y_offset=10):
    """Create stylized chart annotation.

    Args:
        coord: tuple of x and y coordinate
        text: string text for annotation
        label: string label shown on hover
        color: string color. If None (default), will revert to black
        y_offset: offset in yaxis for annotation

    Returns:
        dict: keyword arguments for `go.layout.Annotation`

    """
    return {
        'arrowcolor': color or 'black',
        'arrowhead': 7,
        'arrowsize': 0.3,
        'arrowwidth': 1.5,
        'ax': coord[0],
        'axref': 'x',
        'ay': coord[1] + y_offset,
        'ayref': 'y',
        'bgcolor': color or 'black',
        'bordercolor': color or 'black',
        'borderpad': 2,
        'borderwidth': 1,
        'font': {'color': '#ffffff'},
        'hovertext': label,
        'opacity': 0.8,
        'showarrow': True,
        'text': text,
        'x': coord[0],
        'xref': 'x',
        'y': coord[1],
        'yref': 'y',
    }


class CustomChart:  # noqa: H601
    """Base Class for Custom Charts."""

    annotations = []
    """Store annotations. Default is an empty list.

    See documentation: https://plot.ly/python/reference/#layout-annotations

    """

    _axis_range = {}
    _axis_range_schema = {
        'x': {
            'items': [{'type': ['integer', 'float', 'string']}, {'type': ['integer', 'float', 'string']}],
            'required': False,
            'type': 'list',
        },
        'y': {
            'items': [{'type': ['integer', 'float']}, {'type': ['integer', 'float']}],
            'required': False,
            'type': 'list',
        },
    }

    @property
    def axis_range(self):
        """Specify x/y axis range or leave as empty dictionary for autorange.

        Returns:
            dict: dictionary potentially with keys `(x, y)`

        """
        return self._axis_range

    @axis_range.setter
    def axis_range(self, axis_range):
        errors = validate(axis_range, self._axis_range_schema)
        if errors:
            raise RuntimeError(f'Validation of self.axis_range failed: {errors}')
        # Assign new axis_range
        self._axis_range = axis_range

    def __init__(self, *, title, xlabel, ylabel, layout_overrides=()):
        """Initialize Custom Dash Chart and store parameters as data members.

        Args:
            title: String title for chart (can be an empty string for blank)
            xlabel: XAxis string label (can be an empty string for blank)
            ylabel: YAxis string label (can be an empty string for blank)
            layout_overrides: Custom parameters in format [ParentKey, SubKey, Value] to customize 'go.layout'

        """
        self.title = title
        self.labels = {'x': xlabel, 'y': ylabel}
        self.layout_overrides = layout_overrides
        self.initialize_mutables()

    def initialize_mutables(self):
        """Initialize the mutable data members to prevent modifying one attribute and impacting all instances."""
        ...

    def create_figure(self, df_raw, **kwargs_data):
        """Create the figure dictionary.

        Args:
            df_raw: data to pass to formatter method
            kwargs_data: keyword arguments to pass to the data formatter method

        Returns:
            dict: keys `data` and `layout` for Dash

        """
        return {
            'data': self.create_traces(df_raw, **kwargs_data),
            'layout': go.Layout(self.apply_custom_layout(self.create_layout())),
        }

    def create_traces(self, df_raw, **kwargs_data):
        """Return traces for plotly chart.

        Should return, list: trace data points. List may be empty

        Args:
            df_raw: data to pass to formatter method
            kwargs_data: keyword arguments to pass to the data formatter method

        Raises:
            NotImplementedError: Must be overridden by child class

        """
        raise NotImplementedError('create_traces must be implemented by child class')  # pragma: no cover

    def create_layout(self):
        """Return the standard layout. Can be overridden and modified when inherited.

        Returns:
            dict: layout for Dash figure

        """
        layout = {
            'annotations': self.annotations,
            'title': go.layout.Title(text=self.title),
            'xaxis': {
                'automargin': True,
                'title': self.labels['x'],
            },
            'yaxis': {
                'automargin': True,
                'title': self.labels['y'],
                'zeroline': True,
            },
            'legend': {'orientation': 'h', 'y': -0.25},  # below XAxis label
            'hovermode': 'closest',
        }

        # Optionally apply the specified range
        for axis in ['x', 'y']:
            axis_name = f'{axis}axis'
            if axis in self.axis_range:
                layout[axis_name]['range'] = self.axis_range[axis]
            else:
                layout[axis_name]['autorange'] = True

        return layout

    def apply_custom_layout(self, layout):
        """Extend and/or override layout with custom settings.

        Args:
            layout: base layout dictionary. Typically from self.create_layout()

        Returns:
            dict: layout for Dash figure

        """
        for parent_key, sub_key, value in self.layout_overrides:
            if sub_key is not None:
                layout[parent_key][sub_key] = value
            else:
                layout[parent_key] = value

        return layout


class MarginalChart(CustomChart):  # noqa: H601
    """Base Class for Custom Charts with Marginal X and Marginal Y Plots."""

    def create_figure(self, df_raw, **kwargs_data):
        """Create the figure dictionary.

        Args:
            df_raw: data to pass to formatter method
            kwargs_data: keyword arguments to pass to the data formatter method

        Returns:
            dict: Dash figure object

        """
        # Initialize figure with subplots
        fig = make_subplots(
            rows=2, cols=2,
            shared_xaxes=True, shared_yaxes=True,
            vertical_spacing=0.02, horizontal_spacing=0.02,
            row_width=[0.8, 0.2], column_width=[0.8, 0.2],
        )
        # Populate the traces of each subplot
        traces = [
            (self.create_traces, 2, 1),
            (self.create_marg_top, 1, 1),
            (self.create_marg_right, 2, 2),
        ]
        for trace_func, row, col in traces:
            for trace in trace_func(df_raw, **kwargs_data):
                fig.add_trace(trace, row, col)
        # Apply axis labels
        fig.update_xaxes(title_text=self.labels['x'], row=2, col=1)
        fig.update_yaxes(title_text=self.labels['y'], row=2, col=1)
        # Replace the default blue/white grid introduced in Plotly v4
        fig.update_xaxes(showgrid=True, gridcolor='white')
        fig.update_yaxes(showgrid=True, gridcolor='white')
        fig['layout'].update(self.apply_custom_layout(self.create_layout()))
        return fig

    def create_traces(self, df_raw, **kwargs_data):
        """Return traces for the main plotly chart.

        Should return, list: trace data points. List may be empty

        Args:
            df_raw: data to pass to formatter method
            kwargs_data: keyword arguments to pass to the data formatter method

        Raises:
            NotImplementedError: Must be overridden by child class

        """
        raise NotImplementedError('create_traces must be implemented by child class')  # pragma: no cover

    def create_marg_top(self, df_raw, **kwargs_data):
        """Return traces for the top marginal chart.

        Should return, list: trace data points. List may be empty

        Args:
            df_raw: data to pass to formatter method
            kwargs_data: keyword arguments to pass to the data formatter method

        Raises:
            NotImplementedError: Must be overridden by child class

        """
        raise NotImplementedError('create_marg_top must be implemented by child class')  # pragma: no cover

    def create_marg_right(self, df_raw, **kwargs_data):
        """Return traces for the right marginal chart.

        Should return, list: trace data points. List may be empty

        Args:
            df_raw: data to pass to formatter method
            kwargs_data: keyword arguments to pass to the data formatter method

        Raises:
            NotImplementedError: Must be overridden by child class

        """
        raise NotImplementedError('create_marg_right must be implemented by child class')  # pragma: no cover

    def create_layout(self, *, bg_color='#F0F0F0'):
        """Remove axis lables from base layout as they would be applied to (row=1,col=1).

        Args:
            bg_color: Background color for the chart. Default is white for light themes

        Returns:
            dict: updated layout for Dash figure

        """
        layout = super().create_layout()
        layout['xaxis']['title'] = ''
        layout['yaxis']['title'] = ''
        layout['plot_bgcolor'] = bg_color
        return layout



"""Utilities for generating static HTML reports."""

import io
from base64 import b64encode

import dash_bootstrap_components as dbc
import dominate
import markdown
import pandas as pd
import plotly.io
from bs4 import BeautifulSoup
from dominate import tags, util


def write_div(figure, path_or_file_object, is_div=True, **html_kwargs):
    """Write Plotly figure as HTML to specified file.

    Args:
        figure: Plotly figure (can be from `create_figure` for custom charts)
        path_or_file_object: *string* path or file object
        is_div: if True (default) will override html_kwargs to only write the minimum HTML needed
        html_kwargs: additional keyword arguments passed to `plotly.io.write_html()`

    """
    for key in ['include_plotlyjs', 'full_html']:
        if key not in html_kwargs and is_div:
            html_kwargs[key] = False

    plotly.io.write_html(fig=figure, file=path_or_file_object, **html_kwargs)


def make_div(figure, **html_kwargs):
    """Write Plotly figure as HTML to specified file.

    Args:
        figure: Plotly figure (can be from `create_figure` for custom charts)
        html_kwargs: additional keyword arguments passed to `plotly.io.write_html()`

    Returns:
        str: HTML div

    """
    with io.StringIO() as output:
        write_div(figure, output, is_div=True, **html_kwargs)
        return output.getvalue()


def add_image(image_path, alt_text=None):
    """Write base64 image to HTML.

    Args:
        image_path: Path to image file and format will be read from file suffix.
        alt_text: alternate text. If None, will show the image filename

    Returns:
        str: HTML

    """
    with open(image_path, 'rb') as image_file:
        encoded_image = b64encode(image_file.read()).decode()
    image_uri = f'data:image/{image_path.suffix[1:]};base64,{encoded_image}'
    return f'<img src="{image_uri}" alt="{alt_text or image_path.name}"/>'


def add_video(video_path, alt_text=None):
    """Write base64 video to HTML.

    Video formats can easily be converted with ffmpeg: `ffmpeg -i video_filename.mov video_filename.webm`

    Args:
        video_path: Path to video file and format will be read from file suffix. Video should be in webm format
        alt_text: alternate text. If None, will show the video filename

    Returns:
        str: HTML video tag

    """
    with open(video_path, 'rb') as video_file:
        encoded_video = b64encode(video_file.read()).decode()
    video_uri = f'data:video/{video_path.suffix[1:]};base64,{encoded_video}'
    return f'<video src="{video_uri}" controls>{alt_text or video_path.name}</video>'


def write_image_file(figure, path_or_file_object, image_format, **img_kwargs):
    """Write Plotly figure as an image to specified file.

    Args:
        figure: Plotly figure (can be from `create_figure` for custom charts)
        path_or_file_object: *string* path or file object
        image_format: one of `(png, jpg, jpeg, webp, svg, pdf)`
        img_kwargs: additional keyword arguments passed to `plotly.io.write_image()`

    """
    plotly.io.write_image(fig=figure, file=str(path_or_file_object), format=image_format, **img_kwargs)


def capture_plotly_body():
    """Return HTML body that includes necessary scripts for Plotly and MathJax.

    Returns:
        tuple: of the top and the bottom HTML content

    """
    # Capture necessary Plotly boilerplate HTML
    with io.StringIO() as output:
        write_div({}, output, is_div=False, include_mathjax='.js', validate=False)
        blank_plotly = BeautifulSoup(output.getvalue(), features='lxml')
    # Remove the empty figure div and corresponding script
    plot_div = blank_plotly.find('div', attrs={'class': 'plotly-graph-div'})
    for script in blank_plotly.find_all('script')[::-1]:
        # Use the ID from the plot to identify which script needs to be removed
        if plot_div.attrs['id'] in script.prettify():
            script.decompose()
            break
    plot_div.decompose()
    return blank_plotly.body.prettify()


def format_plotly_boilerplate(**doc_kwargs):
    """Initialize a boilerplate dominate document for creating static Plotly HTML files.

    See dominate documentation: https://pypi.org/project/dominate/

    Args:
        doc_kwargs: keyword arguments for `dominate.document()`

    Returns:
        dict: dominate document instance

    """
    doc = dominate.document(**doc_kwargs)
    with doc:
        util.raw(capture_plotly_body())
    return doc


def create_dbc_doc(theme=dbc.themes.BOOTSTRAP, custom_styles='', **doc_kwargs):
    """Create boilerplate dominate document with Bootstrap and Plotly for static HTML.

    Based on: https://github.com/facultyai/dash-bootstrap-components/tree/master/docs/templates/partials

    See dominate documentation: https://pypi.org/project/dominate/

    Args:
        theme: string URL to CSS for theming Bootstrap. Default is `dbc.themes.BOOTSTRAP`
        custom_styles: optional custom CSS to add to file. Default is blank (`''`)
        doc_kwargs: keyword arguments for `dominate.document()`

    Returns:
        dict: dominate document instance

    """
    stylesheets = [
        {'href': 'https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@9.13.1/build/styles/a11y-light.min.css'},
        {'href': theme},
    ]
    scripts = [
        {'src': 'https://code.jquery.com/jquery-3.4.1.slim.min.js'},
        {'src': 'https://cdn.jsdelivr.net/npm/popper.js@1.16.0/dist/umd/popper.min.js'},
        {'src': 'https://stackpath.bootstrapcdn.com/bootstrap/4.4.1/js/bootstrap.min.js'},
        {'src': 'https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@9.13.1/build/highlight.min.js'},
    ]

    doc = format_plotly_boilerplate(**doc_kwargs)
    with doc.head:
        tags.meta(charset='utf-8')
        tags.meta(name='viewport', content='width=device-width, initial-scale=1')
        for sheet_kwargs in stylesheets:
            tags.link(rel='stylesheet', **sheet_kwargs)
        util.raw(f'<style>{custom_styles}</style>')
        for script_kwargs in scripts:
            tags.script(**script_kwargs)
        util.raw('<script>hljs.initHighlightingOnLoad();</script>')

    return doc


def tag_code(text, language=''):
    """Format HTML for a `pre.code` block with specified `hljs` language.

    Args:
        text: string to show in code block
        language: `hljs` language (ex: `'language-json'`). Default is `''`.

    """
    tags.pre().add(tags.code(text, _class=f'{language} hljs'))


def tag_markdown(text, **markdown_kwargs):
    """Format markdown text as HTML.

    Args:
        text: markdown string
        markdown_kwargs: additional keyword arguments for `markdown.markdown`, such as `extensions`

    """
    util.raw(markdown.markdown(text, **markdown_kwargs))


def tag_table(df_table, table_class=None):
    """Format HTML for a responsive Bootstrap table.

    See Bootstrap documentation at: https://getbootstrap.com/docs/4.4/content/tables/#tables

    Args:
        df_table: pandas dataframe to show in table
        table_class: string classes to add to table. If None, will use default string

    Raises:
        RuntimeError: if `df_table` is not a DataFrame object

    """
    if table_class is None:
        table_class = 'table table-bordered table-striped table-hover'
    if not isinstance(df_table, pd.core.frame.DataFrame):
        raise RuntimeError(f'df_table is not a DataFrame ({type(df_table)}):\n{df_table}')

    df_table = df_table.reset_index()

    with tags.div(_class='table-responsive').add(tags.table(_class=table_class)):
        # Create header row
        with tags.thead().add(tags.tr()):
            for col in df_table.columns:
                tags.th(col)
        # Create body rows
        with tags.tbody():
            for row in df_table.itertuples(index=False):
                with tags.tr():
                    for value in row:
                        tags.td(str(value))


def write_lookup(key, function_lookup):
    """Determine the lookup result and add to the file.

    Args:
        key: string key for function lookup
        function_lookup: dictionary with either the string result or equation and arguments

    Raises:
        RuntimeError: if error in lookup dictionary

    """
    try:
        match = function_lookup[key]
    except KeyError:
        raise RuntimeError(f'Could not find "{key}" in {function_lookup}')

    if isinstance(match, str):
        util.raw(match)
    elif len(match) == 2:
        fun, args = match
        result = fun(*args)
        if isinstance(result, str):
            util.raw(result)
    else:
        raise RuntimeError(f'Match failed for "{key}". Returned: {match} from {function_lookup}')


def markdown_machine(lines, function_lookup):  # noqa: CCR001
    """Convert markdown text file into Plotly-HTML and write to doc context.

    Note: you will need a document with necessary boilerplate and call this within a `with doc:` dominate context
    Multiple Markdown files can then be put into a single HTML output file by calling this function with new lines
    and function lookup arguments

    Args:
        lines: list of text file lines
        function_lookup: dictionary with either the string result or equation and arguments
            Will be inserted into file where `>>lookup:function_name` assuming key of `function_name`

    """
    markdown_lines = []
    for line in lines:
        if line.startswith('>>lookup:'):
            # Write stored markdown and clear list. Then write the matched lookup
            tag_markdown('\n'.join(markdown_lines))
            markdown_lines = []
            write_lookup(line.split('>>lookup:')[1], function_lookup)
        else:
            markdown_lines.append(line)
    if markdown_lines:
        tag_markdown('\n'.join(markdown_lines))


def write_from_markdown(filename, function_lookup, **dbc_kwargs):
    """Wrap markdown_machine to convert markdown to Bootstrap HTML.

    Args:
        filename: path to markdown file
        function_lookup: dictionary with either the string result or equation and arguments
            Will be inserted into file where `>>lookup:function_name` assuming key of `function_name`
        dbc_kwargs: keyword arguments to pass to `create_dbc_doc`

    Returns:
        Path: created HTML filename

    """
    lines = filename.read_text().split('\n')
    html_filename = filename.parent / f'{filename.stem}.html'
    doc = create_dbc_doc(**dbc_kwargs)
    with doc:
        with tags.div(_class='container').add(tags.div(_class='col')):
            markdown_machine(lines, function_lookup)

    html_filename.write_text(str(doc))
    return html_filename



"""Application components built on Dash Bootstrap Components."""

import dash_bootstrap_components as dbc
from dash import dcc


def opts_dd(lbl, value):
    """Format an individual item in a Dash dcc dropdown list.

    Args:
        lbl: Dropdown label
        value: Dropdown value

    Returns:
        dict: keys `label` and `value` for dcc.dropdown()

    """
    return {'label': str(lbl), 'value': value}


def dropdown_group(name, _id, options, form_style=None, **dropdown_kwargs):
    """Return a Form Group with label and dropdown.

    Dropdown documentation: https://dash.plot.ly/dash-core-components/dropdown

    Args:
        name: label name of dropdown
        _id: element id
        options: list of dicts with keys `(value, label)`
        form_style: style keyword argument for dbc.FormGroup(). Default is None
        dropdown_kwargs: key word arguments for dropdown. Could be: `(persistence, multi, searchable, etc.)`

    Returns:
        Row: dbc row with label and dropdown

    """
    if form_style is None:
        form_style = {}
    return dbc.Row(
        [
            dbc.Label(name),
            dcc.Dropdown(id=_id, options=options, **dropdown_kwargs),
        ], style=form_style,
    )


def format_email_pass_id(submit_id):
    """Return tuple of the formatted email and password IDs based on the base submit_id value.

    Args:
        submit_id: id used to create unique element IDs

    Returns:
        tuple: formatted IDs: `(email_id, pass_id)`

    """
    return [f'{submit_id}-{key}' for key in ['email', 'password']]


def login_form(submit_id):
    """Return dbcForm with email and password inputs and submit button.

    Based on: https://dash-bootstrap-components.opensource.faculty.ai/docs/components/form/

    Args:
        submit_id: id used to create unique element IDs

    Returns:
        form: dbc.Form with email and password inputs and submit button

    """
    email_id, pass_id = format_email_pass_id(submit_id)
    return dbc.Form(
        dbc.Row(
            [
                dbc.Label('Email', width='auto'),
                dbc.Col(
                    dbc.Input(type='email', placeholder='Enter email', id=email_id),
                    className='me-3',
                ),
                dbc.Label('Password', width='auto'),
                dbc.Col(
                    dbc.Input(type='password', placeholder='Enter password', id=pass_id),
                    className='me-3',
                ),
                dbc.Col(dbc.Button('Submit', color='primary', id=submit_id), width='auto'),
            ],
            className='g-2',
        ),
    )



"""Utility functions and classes for building applications."""

from copy import deepcopy
from itertools import count
from pathlib import Path
from pprint import pprint
from typing import List, Optional

import dash
from box import Box
from dash import html
from implements import Interface

from .utils_callbacks import format_app_callback

ASSETS_DIR = Path(__file__).parent / 'assets'
"""Path to the static files directory."""

COUNTER = count(start=0, step=1)
"""Initialize iterator to provide set of unique integers when called with `next(...)`."""

STATIC_URLS = {
    'dash': {
        'href': 'https://codepen.io/chriddyp/pen/bWLwgP.css',
        'rel': 'stylesheet',
        'crossorigin': 'anonymous',
    },
    'bulma': {
        'href': 'https://cdnjs.cloudflare.com/ajax/libs/bulma/0.7.5/css/bulma.min.css',
        'rel': 'stylesheet',
        'integrity': 'sha384-H5O3U+oUYwd/bFECZMaQ1XJlueV5e1gB4b7Xt0M06fPbgz48WH33qxUyQNPeZVou',  # pragma: allowlist secret
        'crossorigin': 'anonymous',
    },
    'bulmaswatch-flatly': {
        'href': 'https://jenil.github.io/bulmaswatch/flatly/bulmaswatch.min.css',
        'rel': 'stylesheet',
    },
    'normalize': {
        'href': 'https://cdnjs.cloudflare.com/ajax/libs/normalize/8.0.1/normalize.min.css',
        'rel': 'stylesheet',
        'integrity': 'sha384-9Z9AuAj0Xi0z7WFOSgjjow8EnNY9wPNp925TVLlAyWhvZPsf5Ks23Ex0mxIrWJzJ',  # pragma: allowlist secret
        'crossorigin': 'anonymous',
    },
}
"""Dictionary of stylesheet names and URL to minimized CSS asset.

Hashes generated from: https://www.srihash.org/

"""


def init_app(**app_kwargs: dict) -> dash.Dash:
    """Return new Dash app.

    If not overridden in kwargs, will set path to assets folder, add dash stylesheets, and default meta tags.

    Args:
        app_kwargs: any kwargs to pass to the dash initializer

    Returns:
        dash.Dash: instance (`app`)

    """
    if 'assets_folder' not in app_kwargs:
        app_kwargs['assets_folder'] = str(ASSETS_DIR)
    if 'external_stylesheets' not in app_kwargs:
        app_kwargs['external_stylesheets'] = [STATIC_URLS['dash']]
    if 'meta_tags' not in app_kwargs:
        app_kwargs['meta_tags'] = [{'name': 'viewport', 'content': 'width=device-width, initial-scale=1'}]
    return dash.Dash(__name__, **app_kwargs)


class AppInterface(Interface):  # noqa: H601
    """Base Dash Application Interface."""

    name = None
    _il = {}
    external_stylesheets = []
    modules: list = []
    validation_layout = None
    init_app_kwargs = {}

    def __init__(self, app: Optional[dash.Dash] = None) -> None:  # noqa: D102, D107
        ...

    def create(self, assign_layout: bool = True) -> None:  # noqa: D102
        ...

    def override_module_defaults(self) -> None:  # noqa: D102
        ...

    def initialization(self) -> None:  # noqa: D102
        ...

    def generate_data(self) -> None:  # noqa: D102
        ...

    def register_uniq_ids(self, app_ids: List[str]) -> None:  # noqa: D102
        ...

    def verify_app_initialization(self) -> None:  # noqa: D102
        ...

    def create_elements(self) -> None:
        """Initialize the charts, tables, and other Dash elements.

        Does not return a result. All charts should be initialized in this method (ex `self.chart_main = ParetoChart()`)

        """
        ...

    def return_layout(self) -> dict:  # noqa: D102
        ...

    def callback(self, outputs, inputs, states, pic: bool = False, **kwargs: dict):  # noqa: D102
        ...

    def create_callbacks(self) -> None:
        """Register the chart callbacks.

        Does not return a result. May `pass` as long as no callbacks are needed for application

        """
        ...

    def run(self, **dash_kwargs: dict) -> None:  # noqa: D102
        ...

    def get_server(self):  # noqa: D102
        ...


class AppBase:  # noqa: H601
    """Base class for building Dash Applications."""

    name = None
    """Child class must specify a name for the application.

    Set in `self.__init__()`

    """

    nsi = Box({})
    """Box dictionary of Non-Specific IDs (NSI). In `self.initialization()` automatically registered.

    Set in `self.__init__()`

    """

    _il = {}
    """Specific ID lookup (IL) used to track each element in UI that requires a callback."""

    _id = {}
    """Dotted-dict to store ID's relevant for a given chart (replaces storing each id as a data member).

    FIXME: Need to decide if there is a better approach. Reading this code is confusing...

    """

    external_stylesheets = [STATIC_URLS['dash']]
    """List of external stylesheets. Default is minimal Dash CSS. Only applies if app argument not provided."""

    modules: list = []
    """Initialized modules for GUI set in `self.initialization()`. Leave as an empty list if no modules are needed.

    If list contains modules, in self.create(), each module's `*.create_elements` / `*.create_callbacks` will be called

    Child class must call `*.return_layout(ids)` to render each module's layout in `self.return_layout()` method

    """

    validation_layout = None
    """Validation layout used for callback validation. If None, will use the default layout for callback exceptions."""

    init_app_kwargs = {}
    """Additional keyword arguments passed to `init_app()`."""

    # In child class, declare the rest of the static data members here

    def __init__(self, app: Optional[dash.Dash] = None) -> None:
        """Initialize app and initial data members. Should be inherited in child class and called with super().

        Args:
            app: Dash instance. If None, will create standalone app. Otherwise, will be part of existing app

        """
        default = {'title': self.name, 'external_stylesheets': self.external_stylesheets}
        self.init_app_kwargs = {**default, **self.init_app_kwargs}
        self.app = app or init_app(**self.init_app_kwargs)

    def create(self, assign_layout: bool = True) -> None:  # noqa: CCR001
        """Create the ids, app charts, layout, callbacks, and optional modules.

        Args:
            assign_layout: if True, will assign `self.app.layout`. If False, must call `self.return_layout` separately.
                Default is True

        Raises:
            NotImplementedError: if child class has not set the `self.name` data member

        """
        if self.name is None:  # pragma: no cover
            raise NotImplementedError('Child class must set `self.name` to a unique string for this app')

        # Initialize app and each module
        self.initialization()
        for mod in self.modules:
            self.register_uniq_ids(mod.all_ids)
        self.override_module_defaults()  # Call optional override method

        # Create charts for app and each module
        self.create_elements()
        for mod in self.modules:
            mod.create_elements(self._il)

        # Create app layout. User must call the return_layout method from each module within own return_layout method
        if assign_layout:
            self.app.layout = self.return_layout()
        if assign_layout and self.validation_layout:
            self.app.validation_layout = [deepcopy(self.app.layout)] + self.validation_layout
            pprint('\n\nValidationLayout?')
            pprint(self.app.validation_layout)

        # Create callbacks for app and each module
        self.create_callbacks()
        for mod in self.modules:
            mod.create_callbacks(self)

        self.verify_app_initialization()

    def override_module_defaults(self) -> None:
        """Override default values from modules."""
        ...

    def initialization(self) -> None:
        """Initialize ids with `self.register_uniq_ids([...])` and other one-time actions."""
        self.register_uniq_ids([*self._id.values()])
        self.generate_data()

    def generate_data(self) -> None:
        """Recommended method for generating data stored in memory. Called in initialization."""
        ...

    def register_uniq_ids(self, app_ids: List[str]) -> None:
        """Register the `app_ids` to the corresponding global_id in the `self._il` lookup dictionary.

        The app_ids must be unique within this App so that a layout can be created. This method registers `self._il`
          which are a list of globally unique ids (incorporating this App's unique `self.name`) allowing for the child
          class of this base App to be resused multiple times within a tabbed or multi-page application without
          id-collision

        Args:
            app_ids: list of strings that are unique within this App

        """
        self._il = deepcopy(self._il)
        for app_id in app_ids:
            self._il[app_id] = f'{self.name}-{app_id}'.replace(' ', '-')

    def verify_app_initialization(self) -> None:
        """Check that the app was properly initialized.

        Raises:
            RuntimeError: if child class has not called `self.register_uniq_ids`

        """
        if not self._il.keys():  # pragma: no cover
            raise RuntimeError('Child class must first call `self.register_uniq_ids(__)` before self.run()')

    def return_layout(self) -> dict:
        """Return Dash application layout.

        Returns:
            dict: Dash HTML object. Default is simple HTML text

        """
        return html.Div(['Welcome to the BaseApp! Override return_layout() in child class.'])  # pragma: no cover

    def callback(self, outputs, inputs, states, pic: bool = False, **kwargs: dict):
        """Return app callback decorator based on provided components.

        Args:
            outputs: list of tuples with app_id and property name
            inputs: list of tuples with app_id and property name
            states: list of tuples with app_id and property name
            pic: If True, prevent call on page load (`prevent_initial_call`). Default is False
            **kwargs: any additional keyword arguments for `self.app.callback`

        Returns:
            dict: result of `self.app.callback()`

        """
        return self.app.callback(
            *format_app_callback(self._il, outputs, inputs, states),
            prevent_initial_call=pic,
            **kwargs,
        )

    def run(self, **dash_kwargs: dict) -> None:
        """Launch the Dash server instance.

        Args:
            **dash_kwargs: keyword arguments for `Dash.run_server()`

        """
        self.app.run_server(**dash_kwargs)  # pragma: no cover

    def get_server(self):
        """Retrieve server from app instance for production hosting with green unicorn, waitress, IIS, etc.

        Returns:
            dict: the Flask `server` component of the Dash app

        """
        return self.app.server



"""Upload module and helpers for managing file upload and download.

Some functions based on code from:
https://docs.faculty.ai/user-guide/apps/examples/dash_file_upload_download.html

"""

import base64
import io
import json
import time
from datetime import datetime
from pathlib import Path
from urllib.parse import quote as urlquote

import dash_bootstrap_components as dbc
import pandas as pd
from dash import dash_table, dcc, html

from .utils_app_modules import ModuleBase
from .utils_callbacks import map_args, map_outputs
from .utils_dataset import DBConnect
from .utils_json_cache import CACHE_DIR


def split_b64_file(b64_file):
    """Separate the data type and data content from a b64-encoded string.

    Args:
        b64_file: file encoded in base64

    Returns:
        tuple: of strings `(content_type, data)`

    """
    return b64_file.encode('utf8').split(b';base64,')


def save_file(dest_path, b64_file):
    """Decode and store a file uploaded with Plotly Dash.

    Args:
        dest_path: Path on server filesystem to save the file
        b64_file: file encoded in base64

    """
    data = split_b64_file(b64_file)[1]
    dest_path.write_text(base64.decodebytes(data).decode())


def uploaded_files(upload_dir):
    """List the files in the upload directory.

    Args:
        upload_dir: directory where files are uploadedfolder

    Returns:
        list: Paths of uploaded files

    """
    return [*upload_dir.glob('*.*')]


def file_download_link(filename):
    """Create a Plotly Dash 'A' element that when clicked triggers a file downloaded.

    Args:
        filename: Path to local file to be available for user download

    Returns:
        html.A: clickable Dash link to trigger download

    """
    # PLANNED: Revisit. Should filename be a name or the full path?
    return html.A(filename, href=f'/download/{urlquote(filename)}')


def parse_uploaded_image(b64_file, filename, timestamp):
    """Create an HTML element to show an uploaded image.

    Args:
        b64_file: file encoded in base64
        filename: filename of upload file. Name only
        timestamp: upload timestamp

    Returns:
        html.Img: if image data type

    Raises:
        RuntimeError: if filetype is not a supported image type

    """
    content_type, data = split_b64_file(b64_file)
    if 'image' not in content_type:
        raise RuntimeError(f'Not image type. Found: {content_type}')
    return html.Img(src=b64_file)


def parse_json(raw_json):
    """Return dataframe from JSON formatted in the 'records' orientation.

    Args:
        raw_json: json string

    Returns:
        dataframe: uploaded dataframe parsed from JSON

    Raises:
        RuntimeError: if the JSON file can't be parsed

    """
    dict_json = json.loads(raw_json)
    keys = [*dict_json.keys()]
    if len(keys) != 1:
        raise RuntimeError(
            'Expected JSON with format `{data: [...]}` where `data` could be any key.'
            f'However, more than one key was found: {keys}',
        )
    return pd.DataFrame.from_records(dict_json[keys[0]])


def load_df(decoded, filename):
    """Identify file type and parse the uploaded content into a dataframe.

    Args:
        decoded: string contents/data of the file decoded from the full base64 file
        filename: filename of upload file. Name only

    Returns:
        dataframe: uploaded dataframe parsed from source file

    Raises:
        RuntimeError: if file suffix suffix is unsupported

    """
    suffix = Path(filename).suffix.lower()
    if suffix == '.csv':
        df_upload = pd.read_csv(io.StringIO(decoded.decode('utf-8')))

    elif suffix.startswith('.xl'):
        # xlsx will have 'spreadsheet' in `content_type` but xls will not have anything
        df_upload = pd.read_excel(io.BytesIO(decoded))

    elif suffix == '.json':
        df_upload = parse_json(decoded.decode('utf-8'))

    else:
        raise RuntimeError(f'File type ({suffix}) is unsupported. Expected .csv, .xl*, or .json')

    return df_upload  # noqa: R504


def parse_uploaded_df(b64_file, filename, timestamp):
    """Decode base64 data and parse based on file type. Attempts to return the parsed data as a Pandas dataframe.

    Args:
        b64_file: file encoded in base64
        filename: filename of upload file. Name only
        timestamp: upload timestamp

    Returns:
        dataframe: pandas dataframe parsed from source file

    Raises:
        RuntimeError: if raw data could not be parsed

    """
    content_type, data = split_b64_file(b64_file)
    decoded = base64.b64decode(data)
    try:
        df_upload = load_df(decoded, filename)

    except Exception as error:
        raise RuntimeError(f'Could not parse {filename} ({content_type})\nError: {error}')

    return df_upload  # noqa: R504


def show_toast(message, header, icon='warning', style=None, **toast_kwargs):
    """Create toast notification.

    Args:
        message: string body text
        header: string notification header
        icon: string name in `(primary,secondary,success,warning,danger,info,light,dark)`. Default is warning
        style: style dictionary. Default is the top right
        toast_kwargs: additional toast keyword arguments (such as `duration=5000`)

    Returns:
        dbc.Toast: toast notification from Dash Bootstrap Components library

    """
    if style is None:
        # Position in the top right (note: will occlude the tabs when open, could be moved elsewhere)
        style = {'position': 'fixed', 'top': 10, 'right': 10, 'width': 350, 'zIndex': 1900}
    return dbc.Toast(message, header=header, icon=icon, style=style, dismissable=True, **toast_kwargs)


def drop_to_upload(**upload_kwargs):
    """Create drop to upload element. Dashed box of the active area or a clickable link to use the file dialog.

    Based on dash documentation from: https://dash.plotly.com/dash-core-components/upload

    Args:
        upload_kwargs: keyword arguments for th dcc.Upload element. Children and style are reserved

    Returns:
        dcc.Upload: Dash upload element

    """
    return dcc.Upload(
        children=html.Div(['Drag and Drop or ', html.A('Select a File')]),
        style={
            'width': '100%',
            'height': '60px',
            'lineHeight': '60px',
            'borderWidth': '1px',
            'borderStyle': 'dashed',
            'borderRadius': '5px',
            'textAlign': 'center',
            'margin': '10px',
        },
        **upload_kwargs,
    )


class UploadModule(ModuleBase):  # noqa: H601
    """Module for user data upload.

    Note: this is not intended to be secure

    """

    id_upload = 'upload-drop-area'
    """Unique name for the upload component."""

    id_upload_output = 'upload-output'
    """Unique name for the div to contain output of the parse-upload."""

    id_username_cache = 'username-cache'
    """Unique name for the dcc.Store element to store the current username."""

    all_ids = [id_upload, id_upload_output, id_username_cache]
    """List of ids to register for this module."""

    cache_dir = CACHE_DIR
    """Path to the directory to use for caching files."""

    def __init__(self, *args, **kwargs):
        """Initialize module."""  # noqa: DAR101
        super().__init__(*args, **kwargs)
        self._initialize_database()

    def _initialize_database(self):
        """Create data members `(self.database, self.user_table, self.inventory_table)`."""
        self.database = DBConnect(self.cache_dir / f'_placeholder_app-{self.name}.db')
        self.user_table = self.database.db.create_table(
            'users', primary_id='username', primary_type=self.database.db.types.text,
        )
        self.inventory_table = self.database.db.create_table(
            'inventory', primary_id='table_name', primary_type=self.database.db.types.text,
        )

    def find_user(self, username):
        """Return the database row for the specified user.

        Args:
            username: string username

        Returns:
            dict: for row from table or None if no match

        """
        return self.user_table.find_one(username=username)

    def add_user(self, username):
        """Add the user to the table or update the user's information if already registered.

        Args:
            username: string username

        """
        now = time.time()
        if self.find_user(username):
            self.user_table.upsert({'username': username, 'last_loaded': now}, ['username'])
        else:
            self.user_table.insert({'username': username, 'creation': now, 'last_loaded': now})

    def upload_data(self, username, df_name, df_upload):
        """Store dataframe in database for specified user.

        Args:
            username: string username
            df_name: name of the stored dataframe
            df_upload: pandas dataframe to store

        Raises:
            Exception: If upload fails, deletes the created table

        """
        now = time.time()
        table_name = f'{username}-{df_name}-{int(now)}'
        table = self.database.db.create_table(table_name)
        try:
            table.insert_many(df_upload.to_dict(orient='records'))
        except Exception:
            table.drop()  # Delete the table if upload fails
            raise

        self.inventory_table.insert({
            'table_name': table_name, 'df_name': df_name, 'username': username,
            'creation': now,
        })

    def get_data(self, table_name):
        """Retrieve stored data for specified dataframe name.

        Args:
            table_name: unique name of the table to retrieve

        Returns:
            pd.DataFrame: pandas dataframe retrieved from the database

        """
        table = self.database.db.load_table(table_name)
        return pd.DataFrame.from_records(table.all())

    def delete_data(self, table_name):
        """Remove specified data from the database.

        Args:
            table_name: unique name of the table to delete

        """
        self.database.db.load_table(table_name).drop()

    def return_layout(self, ids):
        """Return the Upload module application layout.

        Args:
            ids: `self._il` from base application

        Returns:
            dict: Dash HTML object.

        """
        return html.Div([
            dcc.Store(id=ids[self.get(self.id_username_cache)], storage_type='session'),
            html.H2('File Upload'),
            html.P('Upload Tidy Data in CSV, Excel, or JSON format'),
            drop_to_upload(id=ids[self.get(self.id_upload)]),
            dcc.Loading(html.Div('', id=ids[self.get(self.id_upload_output)]), type='circle'),
        ])

    def create_callbacks(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        super().create_callbacks(parent)
        self.register_upload_handler(parent)

    def _show_data(self, username):
        """Create Dash HTML to show the raw data loaded for the specified user.

        Args:
            username: string username

        Returns:
            dict: Dash HTML object

        """
        # TODO: Add delete button for each table - need pattern matching callback:
        #   https://dash.plotly.com/pattern-matching-callbacks
        def format_table(df_name, username, creation, raw_df):
            user_str = f'by "{username}" ' if username else ''
            return [
                html.H4(df_name),
                html.P(f'Uploaded {user_str}on {datetime.fromtimestamp(creation)} (Note: only first 10 rows & 10 col)'),
                dash_table.DataTable(
                    data=raw_df[:10].to_dict('records'),
                    columns=[{'name': i, 'id': i} for i in raw_df.columns[:10]],
                    style_cell={
                        'overflow': 'hidden',
                        'textOverflow': 'ellipsis',
                        'maxWidth': 0,
                    },
                ),
                html.Hr(),
            ]

        children = [html.Hr()]
        rows = self.inventory_table.find(username=username)
        for row in sorted(rows, key=lambda _row: _row['creation'], reverse=True):
            df_upload = self.get_data(row['table_name'])
            children.extend(format_table(row['df_name'], row['username'], row['creation'], df_upload))
        return html.Div(children)

    def register_upload_handler(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        outputs = [(self.get(self.id_upload_output), 'children')]
        inputs = [(self.get(self.id_upload), 'contents'), (self.get(self.id_username_cache), 'data')]
        states = [(self.get(self.id_upload), 'filename'), (self.get(self.id_upload), 'last_modified')]

        @parent.callback(outputs, inputs, states, pic=True)
        def upload_handler(*raw_args):
            a_in, a_state = map_args(raw_args, inputs, states)
            b64_file = a_in[self.get(self.id_upload)]['contents']
            username = a_in[self.get(self.id_username_cache)]['data']
            filename = a_state[self.get(self.id_upload)]['filename']
            timestamp = a_state[self.get(self.id_upload)]['last_modified']

            child_output = []
            try:
                if b64_file is not None:
                    df_upload = parse_uploaded_df(b64_file, filename, timestamp)
                    df_upload = df_upload.dropna(axis='columns')  # FIXME: Better handle NaN values...
                    self.add_user(username)
                    self.upload_data(username, filename, df_upload)

            except Exception as error:
                child_output.extend([
                    show_toast(f'{error}', 'Upload Error', icon='danger'),
                    dcc.Markdown(f'### Upload Error\n\n{type(error)}\n\n```\n{error}\n```'),
                ])

            child_output.append(self._show_data(username))

            return map_outputs(outputs, [(self.get(self.id_upload_output), 'children', html.Div(child_output))])



"""Charts for plotting scatter or fitted data."""

import numpy as np
import plotly.graph_objects as go
from scipy import optimize

from .utils_fig import CustomChart, check_raw_data


def create_rolling_traces(df_raw, count_rolling, count_std):
    """Calculate traces for rolling average and standard deviation.

    Args:
        df_raw: pandas dataframe with columns `x: float`, `y: float` and `label: str`
        count_rolling: number of points to use for the rolling calculation
        count_std: number of standard deviations to use for the standard deviation

    Returns:
        list: of Scatter traces for rolling mean and std

    """
    rolling_mean = df_raw['y'].rolling(count_rolling).mean().tolist()
    rolling_std = df_raw['y'].rolling(count_std).std().tolist()
    return [
        go.Scatter(
            fill='toself',
            hoverinfo='skip',
            name=f'{count_std}x STD Range',
            opacity=0.5,
            x=(df_raw['x'].tolist() + df_raw['x'].tolist()[::-1]),
            y=(
                np.add(rolling_mean, np.multiply(count_std, rolling_std)).tolist()
                + np.subtract(rolling_mean, np.multiply(count_std, rolling_std)).tolist()[::-1]
            ),
        ),
        go.Scatter(
            hoverinfo='skip',
            mode='lines',
            name='Rolling Mean',
            opacity=0.9,
            x=df_raw['x'],
            y=rolling_mean,
        ),
    ]


def create_fit_traces(df_raw, name, fit_equation, suppress_fit_errors=False):  # noqa: CCR001
    """Create traces for specified equation.

    Args:
        df_raw: pandas dataframe with columns `name: str`, `x: float`, `y: float` and `label: str`
        name: unique name for trace
        fit_equation: equation used
        suppress_fit_errors: If True, bury errors from scipy fit. Default is False.

    Returns:
        list: of Scatter traces for fitted equation

    """
    fitted_data = []
    try:
        popt, pcov = optimize.curve_fit(fit_equation, xdata=df_raw['x'], ydata=df_raw['y'], method='lm')
        # Calculate representative x values for plotting fit
        x_min = np.min(df_raw['x'])
        x_max = np.max(df_raw['x'])
        x_range = x_max - x_min
        x_values = sorted([
            x_min - 0.05 * x_range,
            *np.divide(range(int(x_min * 10), int(x_max * 10)), 10),
            x_max + 0.05 * x_range,
        ])
        fitted_data = [
            go.Scatter(
                mode='lines+markers',
                name=name,
                opacity=0.9,
                text=f'popt:{[round(param, 3) for param in popt]}',
                x=x_values,
                y=fit_equation(x_values, *popt),
            ),
        ]
    except (RuntimeError, ValueError) as err:  # pragma: no cover
        if not suppress_fit_errors:
            raise

    return fitted_data  # noqa: R504


class RollingChart(CustomChart):
    """Rolling Mean and Filled Standard Deviation Chart for monitoring trends."""

    count_std = 5
    """Count of STD deviations to display. Default 5."""

    count_rolling = count_std
    """Count of items to use for rolling calculations. Default `count_std`."""

    label_data = 'Data'
    """Label for the scatter data. Default is 'Data'."""

    def create_traces(self, df_raw):
        """Return traces for plotly chart.

        Args:
            df_raw: pandas dataframe with columns `x: float`, `y: float` and `label: str`

        Returns:
            list: Dash chart traces

        """
        # Verify data format
        check_raw_data(df_raw, ['x', 'y', 'label'])

        # Create and return the traces
        chart_data = [
            go.Scatter(
                mode='markers',
                name=self.label_data,
                opacity=0.5,
                text=df_raw['label'],
                x=df_raw['x'],
                y=df_raw['y'],
            ),
        ]
        # Only add the rolling calculations if there are a sufficient number of points
        if len(df_raw['x']) >= self.count_rolling:
            chart_data.extend(
                create_rolling_traces(df_raw, self.count_rolling, self.count_std),
            )

        return chart_data


class FittedChart(CustomChart):
    """Scatter Chart with optional Fitted Lines."""

    label_data = 'Data'
    """Label for the scatter data. Default is 'Data'."""

    fit_eqs = []
    """List of fit equations."""

    fallback_mode = 'lines+markers'
    """If not fit_eqs are specified, will fallback to `lines+markers`. Can be set to `markers`."""

    min_scatter_for_fit = 0
    """List of fit equations."""

    suppress_fit_errors = False
    """If True, bury errors from scipy fit and will print message to console. Default is True."""

    def create_traces(self, df_raw):   # noqa: CCR001
        """Return traces for plotly chart.

        Args:
            df_raw: pandas dataframe with columns `name: str`, `x: float`, `y: float` and `label: str`

        Returns:
            list: Dash chart traces

        """
        # Verify data format
        check_raw_data(df_raw, ['name', 'x', 'y', 'label'])

        # Separate raw tidy dataframe into separate scatter plots
        scatter_data = []
        fit_traces = []
        for name in set(df_raw['name']):
            df_name = df_raw[df_raw['name'] == name]
            scatter_data.append(
                go.Scatter(
                    customdata=[name],
                    mode='markers' if self.fit_eqs else self.fallback_mode,
                    name=name,
                    opacity=0.5,
                    text=df_name['label'],
                    x=df_name['x'],
                    y=df_name['y'],
                ),
            )

            if len(df_name['x']) > self.min_scatter_for_fit:
                for fit_name, fit_equation in self.fit_eqs:
                    fit_traces.extend(
                        create_fit_traces(df_name, f'{name}-{fit_name}', fit_equation, self.suppress_fit_errors),
                    )

        return scatter_data + fit_traces



"""R daattali/TimeVis-like charts for displaying chronological data.

See daattali/TimeVis: https://github.com/daattali/timevis

# NOTE: Consider automated (non-overlapping) text/event placement

- MATLAB Adjust Text: https://github.com/Phlya/adjustText
- D3 Labeler: https://github.com/tinker10/D3-Labeler
- Plotly has implementation for contour? https://github.com/plotly/plotly.js/issues/4674#issuecomment-603571483

"""

import numpy as np
import plotly.graph_objects as go

from .utils_data import DASHED_TIME_FORMAT_YEAR, GDP_TIME_FORMAT, format_unix, get_unix
from .utils_fig import CustomChart


class TimeVisChart(CustomChart):  # noqa: H601
    """Time Vis Chart: resource use timeline."""

    date_format = DASHED_TIME_FORMAT_YEAR
    """Date format for bar chart. Default is `DASHED_TIME_FORMAT_YEAR`."""

    fillcolor = '#D5DDF6'
    """Default fillcolor for time vis events."""

    hover_label_settings = {'bgcolor': 'white', 'font_size': 12, 'namelength': 0}
    """Plotly hover label settings."""

    rh = 1
    """Height of each rectangular time vis."""

    y_space = -1.5 * rh
    """Vertical spacing between rectangles."""

    categories = None
    """List of string category names set in self.create_traces()."""

    _shapes = []
    """List of shapes for plotly layout."""

    def create_traces(self, df_raw):   # noqa: CCR001
        """Return traces for plotly chart.

        Args:
            df_raw: pandas dataframe with columns: `(category, label, start, end)`

        Returns:
            list: Dash chart traces

        """
        # Get all unique category names and create lookup for y positions
        self.categories = sorted(cat for cat in set(df_raw['category'].tolist()) if cat)
        y_pos_lookup = {cat: self.y_space * idx for idx, cat in enumerate(self.categories)}
        # Create the Time Vis traces
        traces = []
        self._shapes = []
        self._annotations = []
        for vis in df_raw.itertuples():
            if vis.category in y_pos_lookup:
                y_pos = y_pos_lookup[vis.category]
                if vis.end:
                    traces.append(self._create_time_vis_shape(vis, y_pos))
                    if vis.label:
                        traces.append(self._create_annotation(vis, y_pos))
                else:
                    traces.append(self._create_event(vis, y_pos))
            else:
                y_pos = 0
                traces.append(self._create_non_cat_shape(vis, y_pos))
        return traces

    def _create_hover_text(self, vis):
        """Return hover text for given trace.

        Args:
            vis: row tuple from df_raw with: `(category, label, start, end)`

        Returns:
            string: HTML-formatted hover text

        """
        new_format = f'%a, {GDP_TIME_FORMAT}'
        start_date = format_unix(get_unix(vis.start, self.date_format), new_format)
        if vis.end:
            end_date = format_unix(get_unix(vis.end, self.date_format), new_format)
            date_range = f'<b>Start</b>: {start_date}<br><b>End</b>: {end_date}'
        else:
            date_range = f'<b>Event</b>: {start_date}'
        return f'<b>{vis.category}</b><br>{vis.label}<br>{date_range}'

    def _create_non_cat_shape(self, vis, y_pos):
        """Create non-category time visualization (vertical across all categories).

        Note: background shape is set below a transparent trace so that hover works

        Args:
            vis: row tuple from df_raw with: `(category, label, start, end)`
            y_pos: top y-coordinate of vis

        Returns:
            trace: single Dash chart Scatter trace

        """
        bot_y = self.y_space * len(self.categories)
        self._shapes.append(
            go.layout.Shape(
                fillcolor=self.fillcolor,
                layer='below',
                line={'width': 0},
                opacity=0.4,
                type='rect',
                x0=vis.start,
                x1=vis.end,
                xref='x',
                y0=bot_y,
                y1=y_pos,
                yref='y',
            ),
        )
        return go.Scatter(
            fill='toself',
            opacity=0,
            hoverlabel=self.hover_label_settings,
            line={'width': 0},
            mode='lines',
            text=self._create_hover_text(vis),
            x=[vis.start, vis.end, vis.end, vis.start, vis.start],
            y=[y_pos, y_pos, bot_y, bot_y, y_pos],
        )

    def _create_time_vis_shape(self, vis, y_pos):
        """Create filled rectangle for time visualization.

        Args:
            vis: row tuple from df_raw with: `(category, label, start, end)`
            y_pos: top y-coordinate of vis

        Returns:
            trace: single Dash chart Scatter trace

        """
        return go.Scatter(
            fill='toself',
            fillcolor=self.fillcolor,
            hoverlabel=self.hover_label_settings,
            line={'width': 0},
            mode='lines',
            text=self._create_hover_text(vis),
            x=[vis.start, vis.end, vis.end, vis.start, vis.start],
            y=[y_pos, y_pos, y_pos - self.rh, y_pos - self.rh, y_pos],
        )

    def _create_annotation(self, vis, y_pos):
        """Add vis label to chart as text overlay.

        Args:
            vis: row tuple from df_raw with: `(category, label, start, end)`
            y_pos: top y-coordinate of vis

        Returns:
            trace: single Dash chart Scatter trace

        """
        return go.Scatter(
            hoverlabel=self.hover_label_settings,
            hovertemplate=self._create_hover_text(vis) + '<extra></extra>',
            hovertext=self._create_hover_text(vis),
            mode='text',
            text=vis.label,
            textposition='middle right',
            x=[vis.start],
            y=[y_pos - self.rh / 2],
        )

    def _create_event(self, vis, y_pos):
        """Create singular event with vertical line, marker, and text.

        If label is longer than 10 characters, then the annotation is shown offset with an arrow.

        Args:
            vis: row tuple from df_raw with: `(category, label, start, end)`
            y_pos: top y-coordinate of vis

        Returns:
            trace: single Dash chart Scatter trace

        """
        if len(vis.label) > 10:
            self._annotations.append({
                'align': 'right',
                'arrowcolor': self.fillcolor,
                'showarrow': True,
                'arrowhead': 2,
                'text': vis.label,
                'x': vis.start,
                'xanchor': 'right',
                'y': y_pos - self.rh / 2,
                'yanchor': 'middle',
            })
        self._shapes.append(
            go.layout.Shape(
                layer='below',
                line={
                    'color': self.fillcolor,
                    'dash': 'longdashdot',
                    'width': 2,
                },
                type='line',
                x0=vis.start,
                x1=vis.start,
                xref='x',
                y0=self.y_space * len(self.categories),
                y1=y_pos - self.rh / 2,
                yref='y',
            ),
        )
        return go.Scatter(
            hoverlabel=self.hover_label_settings,
            hovertemplate=self._create_hover_text(vis) + '<extra></extra>',
            hovertext=self._create_hover_text(vis),
            marker={'color': self.fillcolor},
            mode='markers+text',
            text='' if len(vis.label) > 10 else vis.label,
            textposition='top center',
            x=[vis.start],
            y=[y_pos - self.rh / 2],
        )

    def create_layout(self):
        """Extend the standard layout.

        Returns:
            dict: layout for Dash figure

        """
        layout = super().create_layout()
        # Set YAxis tick marks for category names (https://plotly.com/python/tick-formatting)
        layout['yaxis']['tickmode'] = 'array'
        layout['yaxis']['tickvals'] = np.subtract(
            np.multiply(
                np.array(range(len(self.categories))),
                self.y_space,
            ),
            self.rh / 2,
        )
        layout['yaxis']['ticktext'] = [*self.categories]
        layout['yaxis']['zeroline'] = False
        # Hide legend
        layout['legend'] = {}
        layout['showlegend'] = False
        # Add shapes and append new annotations
        layout['shapes'] = self._shapes
        layout['annotations'] += self._annotations
        return layout



"""Utilities to build modules (delegated layout & callback methods) for Dash apps."""

import pandas as pd
from dash import dcc


class ModuleBase:  # noqa: H601
    """Base class for building a modular component for use in a Dash application."""

    all_ids = None
    """List of ids to register for this module."""

    def __init__(self, name):
        """Initialize module.

        Args:
            name: unique string name for this module

        Raises:
            NotImplementedError: if child class has not created a list, `all_ids`

        """
        if self.all_ids is None:
            raise NotImplementedError('Child class must create list of `self.all_ids`')  # pragma: no cover

        # Make ids unique and update all ids so that modules can be reused in the same app
        self.name = name
        self._ids = {_id: f'{self.name}__{_id}' for _id in self.all_ids}
        self.all_ids = self._ids.values()

        self.initialize_mutables()

    def get(self, _id):
        """Return the  the callback for creating the main chart.

        Args:
            _id: id from this module that is found in `self.all_ids`

        Returns:
            str: unique id name from instance of this module

        """
        return self._ids[_id]

    def initialize_mutables(self):
        """Initialize the mutable data members to prevent modifying one attribute and impacting all instances."""
        ...

    def create_elements(self, ids):
        """Register the callback for creating the main chart.

        Args:
            ids: `self._il` from base application

        """
        ...  # pragma: no cover

    def return_layout(self, ids):
        """Return Dash application layout.

        Args:
            ids: `self._il` from base application

        Raises:
            NotImplementedError: Dash HTML object. Default is simple HTML text

        """
        raise NotImplementedError('Must be implemented')  # pragma: no cover

    def create_callbacks(self, parent):
        """Register callbacks to handle user interaction.

        Args:
            parent: parent instance (ex: `self`)

        """
        ...  # pragma: no cover


class DataCache(ModuleBase):  # noqa: H601
    """Module to store data in UI and later loaded as needed."""

    id_cache = 'cache'
    """Session ID."""

    all_ids = [id_cache]
    """List of ids to register for this module."""

    def return_layout(self, ids, storage_type='memory', **store_kwargs):
        """Return Dash application layout.

        `dcc.Store` documentation: https://dash.plotly.com/dash-core-components/store

        Args:
            ids: `self._il` from base application
            storage_type: `dcc.Store` storage type. Default is memory to clear on refresh
            store_kwargs: additional keyword arguments to pass to `dcc.Store`

        Returns:
            dict: Dash HTML object.

        """
        return dcc.Store(id=ids[self.get(self.id_cache)], storage_type=storage_type, **store_kwargs)

    def return_write_df_map(self, df_table):
        """Return list of tuples for `map_outputs` that includes the new data cache JSON.

        Args:
            df_table: dataframe to show in table

        Returns:
            list: list of tuples for `map_outputs`

        """
        return [(self.get(self.id_cache), 'data', df_table.to_json())]

    def read_df(self, args):
        """Return list of tuples for `map_outputs` that includes the new data cache JSON.

        Args:
            args: either `a_in` or `a_state`, whichever has the id_cache-data

        Returns:
            dataframe: returns dataframe read from `dcc.Store`

        """
        return pd.read_json(args[self.get(self.id_cache)]['data'])



"""Helpers for managing a generic JSON data file cache. Can be used to reduce API calls, etc.

Full dataset documentation: https://dataset.readthedocs.io/en/latest/api.html

"""

import json
import time
from pathlib import Path

from .utils_data import uniq_table_id, write_pretty_json
from .utils_dataset import DBConnect

#  FIXME: Add versioning to the cache directory with semver logic: https://pypi.org/project/semantic-version/

CACHE_DIR = Path(__file__).parent / 'local_cache'
"""Path to folder with all downloaded responses from Kitsu API."""

FILE_DATA = DBConnect(CACHE_DIR / '_file_lookup_database.db')
"""Global instance of the DBConnect() for the file lookup database."""

CACHE_TABLE_NAME = 'files'
"""Table name containing the cache file information."""

ID_KEY = 'identifier'
"""Name of the SQLite column containing the unique identifier."""
TS_KEY = 'timestamp'
"""Name of the SQLite column containing the timestamp."""
FILENAME_KEY = 'filename'
"""Name of the SQLite column containing the string filename."""

# TODO: Enable versioning of data and automatic deletion when the version changes
DATA_VERSION_KEY = 'data_version'
"""Key to indicate the data version."""


def get_files_table(db_instance):
    """Retrieve stored object from cache database.

    Args:
        db_instance: Connected Database file with `DBConnect()`.

    Returns:
        table: Dataset table for the files lookup

    """
    return db_instance.db.load_table(CACHE_TABLE_NAME)


def initialize_cache(db_instance):
    """Ensure that the directory and database exist. Remove files from database if manually removed.

    Args:
        db_instance: Connected Database file with `DBConnect()`.

    """
    table = db_instance.db.create_table(CACHE_TABLE_NAME)

    removed_files = []
    for row in table:
        if not Path(row[FILENAME_KEY]).is_file():
            removed_files.append(row[FILENAME_KEY])

    for filename in removed_files:
        table.delete(filename=filename)


def get_cache_dict(db_instance):
    """Return a dictionary `{identifier: path}` keys and values.

    Args:
        db_instance: Connected Database file with `DBConnect()`.

    Returns:
        dict: dictionary `{identifier: path}` keys and values

    """
    table = get_files_table(db_instance)
    return {row[ID_KEY]: Path(row[FILENAME_KEY]) for row in table}


def match_identifier_in_cache(identifier, db_instance):
    """Return list of matches for the given identifier in the file database.

    Args:
        identifier: identifier to use as a reference if the corresponding data is already cached
        db_instance: Connected Database file with `DBConnect()`.

    Returns:
        list: list of match object with keys of the SQL table

    """
    kwargs = {ID_KEY: identifier}
    return [*get_files_table(db_instance).find(**kwargs)]


def store_cache_as_file(prefix, identifier, db_instance, cache_dir=CACHE_DIR, suffix='.json'):
    """Store the reference in the cache database and return the file so the user can handle saving the file.

    Args:
        prefix: string used to create more recognizable filenames
        identifier: identifier to use as a reference if the corresponding data is already cached
        db_instance: Connected Database file with `DBConnect()`.
        cache_dir: path to the directory to store the file. Default is `CACHE_DIR
        suffix: string filename suffix. The default is `.json`

    Returns:
        Path: to the cached file. Caller needs to write to the file

    Raises:
        RuntimeError: if duplicate match found when storing

    """
    # Check that the identifier isn't already in the database
    matches = match_identifier_in_cache(identifier, db_instance)
    if matches:
        raise RuntimeError(f'Already have an entry for this identifier (`{identifier}`): {matches}')
    # Update the database and store the file
    filename = cache_dir / f'{prefix}_{uniq_table_id()}{suffix}'
    new_row = {FILENAME_KEY: str(filename), ID_KEY: identifier, TS_KEY: time.time()}
    get_files_table(db_instance).insert(new_row)
    return filename


def store_cache_object(prefix, identifier, obj, db_instance, cache_dir=CACHE_DIR):
    """Store the object as a JSON file and track in a SQLite database to prevent duplicates.

    Args:
        prefix: string used to create more recognizable filenames
        identifier: identifier to use as a reference if the corresponding data is already cached
        obj: JSON object to write
        db_instance: Connected Database file with `DBConnect()`.
        cache_dir: path to the directory to store the file. Default is `CACHE_DIR

    Raises:
        Exception: if duplicate match found when storing

    """
    filename = store_cache_as_file(prefix, identifier, db_instance, cache_dir)
    try:
        write_pretty_json(filename, obj)
    except Exception:
        # If writing the file fails, ensure that the record is removed from the database
        get_files_table(db_instance).delete(filename=filename)
        raise


def retrieve_cache_fn(identifier, db_instance):
    """Retrieve stored object from cache database.

    Args:
        identifier: identifier to use as a reference if the corresponding data is already cached
        db_instance: Connected Database file with `DBConnect()`.

    Returns:
        Path: to the cached file. Caller needs to read the file

    Raises:
        RuntimeError: if not exactly one match found

    """
    matches = match_identifier_in_cache(identifier, db_instance)
    if len(matches) != 1:
        raise RuntimeError(f'Did not find exactly one entry for this identifier (`{identifier}`): {matches}')
    return Path(matches[0][FILENAME_KEY])


def retrieve_cache_object(identifier, db_instance):
    """Retrieve stored object from cache database.

    Args:
        identifier: identifier to use as a reference if the corresponding data is already cached
        db_instance: Connected Database file with `DBConnect()`.

    Returns:
        dict: object stored in the cache

    """
    filename = retrieve_cache_fn(identifier, db_instance)
    return json.loads(filename.read_text())



"""Helpers for building Dash applications."""

import csv
import json
import sqlite3
import time
from contextlib import ContextDecorator
from datetime import datetime
from pathlib import Path

import pandas as pd
from cerberus import Validator

# ----------------------------------------------------------------------------------------------------------------------
# For Working with Data


def enable_verbose_pandas(max_columns=None, max_rows=None, max_seq_items=None):
    """Update global pandas configuration for printed dataframes.

    Args:
        max_columns: the number of max columns. Default is None (to show all)
        max_rows: the number of max rows. Default is None (to show all)
        max_seq_items: the number of max sequence items. Default is None (to show all) # TODO: what does this set?

    """
    # Enable all columns to be displayed at once (or tweak to set a new limit)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    pd.set_option('display.max_colwidth', max_columns)

    # Optionally modify number of rows shown
    pd.set_option('display.max_rows', max_rows)
    if max_seq_items:
        pd.options('display.max_seq_items', max_seq_items)


def append_df(df_old, df_new):
    """Handle appending a dataframe if the old_df is None. Useful for iteration.

    Args:
        df_old: dataframe or None
        df_new: new dataframe to append. Expects all columns to match

    Returns:
        dataframe: combined dataframe

    """
    return df_new if df_old is None else pd.concat([df_old, df_new]).reset_index(drop=True)


def validate(document, schema, **validator_kwargs):
    """Validate a data structure. Return errors if any found.

    Cerberus Documentation: https://docs.python-cerberus.org/en/stable/validation-rules.html

    Args:
        document: data structure to validate
        schema: expected structure
        validator_kwargs: additional keyword arguments for Validator class

    Returns:
        list: validation errors

    """
    validator = Validator(schema, **validator_kwargs)
    validator.validate(document)
    return validator.errors


def json_dumps_compact(data):   # noqa: CCR001
    """Format provided dictionary into compact JSON. Lists will be in one line rather than split on new lines.

    Args:
        data: JSON-serializable dictionary

    Returns:
        str: JSON-formatted string with lists compacted into a single line

    """
    clean_data = {}
    # Check each key/value pair to determine if any intermediary strings are needed for later formatting
    for key, raw in data.items():
        # PLANNED: Convert to FP and recursive calls?
        if isinstance(raw, list):
            values = [f'``{value}``' if isinstance(value, str) else value for value in raw]
            clean_data[key] = '[' + ','.join(map(str, values)) + ']'
        else:
            clean_data[key] = raw
    # Format the dictionary into JSON and replace the special characters used as intermediaries
    raw_json = json.dumps(clean_data, indent=4, separators=(',', ': '), sort_keys=True)
    return (
        raw_json
        .replace(': "[', ': [')
        .replace(']"', ']')
        .replace('``', '"')
        .replace("'", '"')
    )


def write_pretty_json(filename, obj):
    """Write indented JSON file.

    Args:
        filename: Path or plain string filename to write (should end with `.json`)
        obj: JSON object to write

    """
    Path(filename).write_text(json.dumps(obj, indent=4, separators=(',', ': ')))


def write_csv(csv_path, rows):
    """Write a csv file with appropriate line terminator and encoding.

    Args:
        csv_path: path to CSV file
        rows: list of lists to write to CSV file

    """
    with open(csv_path, 'w', newline='\n', encoding='utf-8') as csv_file:
        writer = csv.writer(csv_file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        for row in rows:
            writer.writerow(row)


# ----------------------------------------------------------------------------------------------------------------------
# Time Helpers

US_TIME_FORMAT = '%m/%d/%Y %H:%M:%S'
"""String time format with month/year (MM/DD/YYYY HH:MM:SS)."""

DASHED_TIME_FORMAT_US = '%m-%d-%Y %H:%M:%S'
"""Dashed time format with month first (MM-DD-YYYY HH:MM:SS)."""

DASHED_TIME_FORMAT_YEAR = '%Y-%m-%d %H:%M:%S'
"""Dashed time format with year first (YYYY-MM-DD HH:MM:SS)."""

TIME_FORMAT_FILE = '%Y-%m-%d_%H%M%S'
"""Filename-safe time format with year first (YYYY-MM-DD_HHMMSS)."""

GDP_TIME_FORMAT = '%d%b%Y %H:%M:%S'
"""Good Documentation Practice time format (DDMMMYYYY HH:MM:SS)."""


def get_unix(str_ts, date_format):
    """Get unix timestamp from a string timestamp in date_format.

    Args:
        str_ts: string timestamp in `date_format`
        date_format: datetime time stamp format

    Returns:
        int: unix timestamp

    """
    return datetime.strptime(str_ts, date_format).timestamp()


def format_unix(unix_ts, date_format):
    """Format unix timestamp as a string timestamp in date_format.

    Args:
        unix_ts: unix timestamp
        date_format: datetime time stamp format

    Returns:
        string: formatted timestamp in `date_format`

    """
    return datetime.fromtimestamp(unix_ts).strftime(date_format)


# ----------------------------------------------------------------------------------------------------------------------
# General SQL


def uniq_table_id():
    """Return a unique table ID based on the current time in ms.

    Returns:
        str: in format `U<timestamp_ns>`

    """
    return f'U{time.time_ns()}'


# ----------------------------------------------------------------------------------------------------------------------
# sqlite3


class SQLConnection(ContextDecorator):
    """Ensure the SQLite connection is properly opened and closed."""

    def __init__(self, db_path):
        """Initialize context wrapper.

        Args:
            db_path: Path to a SQLite file

        """
        self.conn = None
        self.db_path = db_path

    def __enter__(self):
        """Connect to the database and return connection reference.

        Returns:
            dict: connection to sqlite database

        """
        self.conn = sqlite3.connect(self.db_path)
        return self.conn

    def __exit__(self, exc_type, exc_value, traceback):
        """Close connection."""  # noqa: DAR101
        self.conn.close()


def list_sql_tables(db_path):
    """Return all table names from the SQL database.

    Args:
        db_path: path to SQLite database file

    Returns:
        list: of unique table names in the SQL database

    """
    with SQLConnection(db_path) as conn:
        cursor = conn.cursor()
        cursor.execute('SELECT name FROM sqlite_master WHERE TYPE = "table"')
        return [names[0] for names in cursor.fetchall()]
